import unittest

from player import Player, Participants


class TestParticipants(unittest.TestCase):

    def test_participants_generateNames(self):
        names = Participants.generateNames(3)
        self.assertEqual(names, ['p00', 'p01', 'p02'])

    def test_participants_generageNamesWithCustomPrefix(self):
        names = Participants.generateNames(2, 'prefix')
        self.assertEqual(names, ['prefix00', 'prefix01'])

    def test_participantsCreateFromNames(self):
        names = ['like', 'aman', 'fantomas']
        people = Participants.createFromNames(names)
        self.assertEqual(len(people), 3)

        self.assertEqual(people[0].id, 0)
        self.assertEqual(people[0].name, 'like')
        self.assertEqual(people[1].id, 1)
        self.assertEqual(people[1].name, 'aman')
        self.assertEqual(people[2].id, 2)
        self.assertEqual(people[2].name, 'fantomas')

    def test_participantsCreateFromPlayers(self):
        player0 = Player(525, 'like')
        player1 = Player(727, 'aman')
        player2 = Player(123, 'fantomas')
        players = [player0, player1, player2]

        people = Participants(players)
        self.assertEqual(len(people), 3)

        self.assertEqual(people[0], player0)
        self.assertEqual(people[1], player1)
        self.assertEqual(people[2], player2)

    def test_participants_find_ok(self):
        player0 = Player(525, 'like')
        player1 = Player(727, 'aman')
        player2 = Player(123, 'fantomas')
        players = [player0, player1, player2]
        people = Participants(players)

        self.assertEqual(people.find(525), player0)
        self.assertEqual(people.find(123), player2)
        self.assertEqual(people.find(727), player1)

    def test_participants_find_fail(self):
        player0 = Player(525, 'like')
        player1 = Player(727, 'aman')
        player2 = Player(123, 'fantomas')
        players = [player0, player1, player2]
        people = Participants(players)

        p = people.find(100)
        self.assertEqual(p, None)

    pass


if __name__ == '__main__':
    unittest.main()



import dataclasses

from .player import *


# forward declarations
class Game:
    pass


class GameSet:
    pass


@dataclasses.dataclass()
class Game:
    '''
    Game is played by exactly 10 players.
    We store player IDs, not players.
    '''

    id: int
    players: list[int] = dataclasses.field(
        default_factory=list, hash=False, compare=False)

    @staticmethod
    def create(gs: GameSet) -> Game:
        '''Creates a Game from GameSet'''
        players_list = [id for id in gs.players]
        return Game(gs.id, players_list)

    def isValid(self) -> bool:
        # game must have exactly 10 players
        if len(self.players) != 10:
            return False

        # all players must be unique (by id)
        uniquePlayers = set()
        for playerId in self.players:
            if playerId in uniquePlayers:
                return False
            uniquePlayers.add(playerId)
        return True

    @staticmethod
    def fromJson(d: dict):
        players = [playerId for playerId in d['players']]
        return Game(d['id'], players)


class GameSet:
    '''
    Game for opponents optimization, just set of players, without seat numbers
    '''

    id: int
    players: set[int] = {}

    def __init__(self, id: int, players: set):
        self.id = id
        self.players = players

    @staticmethod
    def create(game: Game) -> GameSet:
        '''Creates a GameSet from Game'''
        player_set = {id for id in game.players}
        return GameSet(game.id, player_set)



from typing import Callable

from .schedule import Schedule
from .metrics import Metrics


class Print:
    @staticmethod
    def print(func: Callable):
        for line in func:
            print(line)

    @staticmethod
    def slots(schedule: Schedule):
        yield ''
        yield "*** Slots:"
        for gameId, slot in schedule.slots.items():
            ids = [player for player in slot.players]
            s = [f"{item:3d}" for item in ids]
            str = ''.join(s)
            yield f"Slot {gameId:2d}: {str}"

    @staticmethod
    def opponentsMatrix(schedule: Schedule):
        m = Metrics(schedule)
        matrix = m.calcOpponentsMatrix()

        yield ''
        yield "*** Opponents matrix:"
        for playerId in range(len(matrix)):
            line = matrix[playerId]
            s = ''.join([f"{v:3d}" for v in line])
            yield f"{playerId:2d}: {s}"

    @staticmethod
    def pairsMatrix(schedule: Schedule):
        m = Metrics(schedule)

        yield ''
        yield "Pairs matrix:"
        for playerId in range(schedule.numPlayers):
            pairs = m.calcPlayerPairsHistogram(playerId)
            penalty = m.penaltyPlayer(playerId)
            header = f"Player {playerId:2d}: err={penalty:6.2f}: "
            str = ''.join([f"{numPairs:4d}" for numPairs in pairs.values()])

            yield f"{header:25s}{str}"

        # ideal line
        hist = m.penaltyIdealHistogram()
        header = f"*** Ideal: "
        str = ''.join([f"{round(numPairs):4d}" for numPairs in hist.values()])
        yield f"{header:25s}{str}"

        # last line
        allPairs = m.calcPairsHistogram()
        header = f"*** Overall: "
        str = ''.join([f"{numPairs:4d}" for numPairs in allPairs.values()])
        yield f"{header:25s}{str}"

    def minMaxPairs(schedule: Schedule, numGames: list[int]):
        m = Metrics(schedule)

        yield ''
        yield "*** Min-Max opponents:"
        for playerId in range(schedule.numPlayers):
            pairs = m.calcPlayerPairs(playerId)

            header = f"Player {playerId:2d}: "
            str = ""
            for idx in numGames:
                if len(pairs[idx]) == 0:
                    continue
                s = ''.join([f"{id:2d} " for id in pairs[idx]]
                            ) if len(pairs[idx]) < 5 else "..."
                str += f"g = {idx} with: [{s}]; "
            yield f"{header}{str}"

    @staticmethod
    def pairsHistogram(schedule: Schedule):
        m = Metrics(schedule)
        allPairs = m.calcPairsHistogram()

        yield ''
        yield "Pairs histogram:"
        for numGames, numPairs in allPairs.items():
            if numPairs > 0:
                yield f"{numGames:2d} : {numPairs:3d} pairs"

    @staticmethod
    def scheduleByGames(schedule: Schedule):
        yield ''
        yield "Schedule by games:"
        for round in schedule.rounds:
            yield f"\nRound: {round.id}"
            for gameId in round.gameIds:
                game = schedule.games[gameId]
                s = [f"{id:3d}" for id in game.players]
                str = ''.join(s)
                yield f"Game {game.id:2d}: {str}"

    @staticmethod
    def scheduleByPlayers(schedule: Schedule):
        yield ''
        yield "Schedule by players:"
        for playerId in range(schedule.numPlayers):
            str = ""
            for round in schedule.rounds:
                roundStr = f"{' *:* '}"
                for table, gameId in enumerate(round.gameIds):
                    game = schedule.games[gameId]
                    tableStr = chr(ord('A') + table)
                    for seat, id in enumerate(game.players):
                        if id == playerId:
                            roundStr = f" {tableStr}:{(seat+1):<2d}"
                            break
                str += roundStr

            if schedule.participants is None:
                header = f"Player {playerId:2d}: "
            else:
                playerName = schedule.participants[playerId].name
                header = f"{playerName:20}:"
            yield f"{header}{str}"

    @staticmethod
    def seatsMatrix(schedule: Schedule):
        m = Metrics(schedule)
        matrix = m.calcSeatsMatrix()

        yield ''
        yield "*** Seats matrix:"
        for playerId in range(len(matrix)):
            line = matrix[playerId]
            s = ''.join([f"{v:3d}" for v in line])
            yield f"{playerId:2d}: {s}"

    @staticmethod
    def mwtSchedule(schedule: Schedule):
        for round in schedule.rounds:
            for seat in range(10):
                ids = [schedule.games[gameId].players[seat]
                       for gameId in round.gameIds]

                # output player IDs or player names
                if not schedule.participants:
                    line = [f"{id:>3d}," for id in ids]
                else:
                    line = [
                        f" {schedule.participants.find(id).name}," for id in ids]

                # print without last comma
                str = ''.join(line)
                yield str[:-1]
            yield ''




from .configuration import Configuration
from .schedule import *
from .round import *
from .game import *
from .player import *


class ScheduleFactory:
    def createInitialSchedule(conf: Configuration):
        if not conf.isValid():
            return None

        # create rounds and games
        rounds = []
        games = []
        gameId = 0
        playerId = 0
        for roundNum in range(conf.numRounds):
            gamesInRound = []

            for tableNum in range(conf.numTables):
                if gameId >= conf.numGames:
                    break

                # prepare players for game
                playerIds = [-1] * 10
                for i in range(10):
                    playerIds[i] = playerId
                    playerId = (playerId + 1) % conf.numPlayers

                # create next game
                game = Game(gameId, playerIds)
                games.append(game)
                gamesInRound.append(gameId)
                gameId = gameId + 1

            # create next round
            round = Round(roundNum, gamesInRound)
            rounds.append(round)

        schedule = Schedule(conf, rounds, games)
        return schedule



import dataclasses


@dataclasses.dataclass(frozen=True, order=False)
class Round:
    '''
    Round is set of games played at the same time.
    Therefore, any player can be only in one of its games.
    '''
    id: int
    gameIds: list[int] = dataclasses.field(
        hash=False, compare=False, default_factory=list)

    @staticmethod
    def fromJson(d: dict):
        gameIds = [id for id in d['gameIds']]
        return Round(d['id'], gameIds)



import dataclasses
import unittest
import json

from mafia_schedule.configuration import Configuration
from mafia_schedule.schedule_factory import ScheduleFactory
from mafia_schedule.schedule import Schedule
from mafia_schedule.round import Round
from mafia_schedule.game import Game
from mafia_schedule.player import Player, Participants


class TestSerialization(unittest.TestCase):
    def test_configuration_toJson(self):
        conf = Configuration(numPlayers=25, numTables=2,
                             numRounds=10, numGames=20, numAttempts=8)
        d = dataclasses.asdict(conf)
        self.assertEqual(d, {'numPlayers': 25,
                             'numTables': 2,
                             'numRounds': 10,
                             'numGames': 20,
                             'numAttempts': 8})

    def test_configuration_fromJson(self):
        d = {'numPlayers': 25,
             'numTables': 2,
             'numRounds': 10,
             'numGames': 20,
             'numAttempts': 8}
        conf = Configuration(**d)
        self.assertEqual(conf.numPlayers, 25)
        self.assertEqual(conf.numTables, 2)
        self.assertEqual(conf.numRounds, 10)
        self.assertEqual(conf.numGames, 20)
        self.assertEqual(conf.numAttempts, 8)

    def test_player_toJson(self):
        player = Player(525, 'like')
        d = dataclasses.asdict(player)
        self.assertEqual(d, {'id': 525, 'name': 'like'})

    def test_player_fromJson(self):
        d = {'id': 525, 'name': 'like'}
        player = Player(**d)
        self.assertEqual(player.id, 525)
        self.assertEqual(player.name, 'like')

    def test_participants_toJson(self):
        participants = Participants.create(3, "Player_")
        d = participants.toJson()
        self.assertEqual(d, {'people': [
            {'id': 0, 'name': 'Player_00'},
            {'id': 1, 'name': 'Player_01'},
            {'id': 2, 'name': 'Player_02'}]
        })

    def test_participants_fromJson(self):
        d = {'people': [
            {'id': 0, 'name': 'Player_0'},
            {'id': 1, 'name': 'Player_1'},
            {'id': 2, 'name': 'Player_2'}]
        }

        p = Participants.fromJson(d)
        self.assertEqual(p[0].id, 0)
        self.assertEqual(p[0].name, 'Player_0')
        self.assertEqual(p[1].id, 1)
        self.assertEqual(p[1].name, 'Player_1')
        self.assertEqual(p[2].id, 2)
        self.assertEqual(p[2].name, 'Player_2')

    def test_game_toJson(self):
        ids = [5, 9, 3, 6, 1, 4, 7, 8, 2, 0]
        game = Game(525, ids)
        d = dataclasses.asdict(game)
        expected = {'id': 525, 'players': ids}
        self.assertEqual(d, expected)

    def test_game_fromJson(self):
        playerLike = Player(112, 'like')
        playerAman = Player(113, 'aman')
        d = {'id': 525,
             'players': [112, 113]}

        game = Game.fromJson(d)
        self.assertEqual(game.id, 525)
        self.assertEqual(len(game.players), 2)
        self.assertEqual(game.players[0], playerLike.id)
        self.assertEqual(game.players[1], playerAman.id)

    def test_round_toJson(self):
        participants = Participants.create(10)
        players = participants.people
        game1 = Game(525, players)
        games = [525]
        round = Round(727, games)
        d = dataclasses.asdict(round)
        self.assertEqual(
            d, {'id': 727, 'gameIds': [525]})

    def test_round_fromJson(self):
        participants = Participants.create(10)
        players = participants.people
        game1 = Game(525, players)
        game2 = Game(727, players)
        gameIds = [game1.id, game2.id]

        d = {'id': 100, 'gameIds': [525, 727]}
        round = Round.fromJson(d)

        self.assertEqual(round.id, 100)
        self.assertEqual(len(round.gameIds), 2)
        self.assertEqual(round.gameIds[0], 525)
        self.assertEqual(round.gameIds[1], 727)

    def test_schedule_toJson(self):
        numPlayers = 10
        conf = Configuration(numPlayers, numTables=1,
                             numRounds=1, numGames=1, numAttempts=1)
        schedule = ScheduleFactory.createInitialSchedule(conf)
        d = schedule.toJson()
        self.assertEqual(d, {'configuration': dataclasses.asdict(conf),
                             'rounds': [dataclasses.asdict(round) for round in schedule.rounds],
                             'games': [dataclasses.asdict(game) for game in schedule.games]})

    pass


if __name__ == '__main__':
    unittest.main()



import unittest

from configuration import Configuration
from schedule_factory import ScheduleFactory
from schedule import Schedule


class TestScheduleFactory(unittest.TestCase):
    def test_factory_simple(self):
        '''
        Simple schedule: 10 players, just 1 game
        '''
        conf = Configuration(numPlayers=10, numTables=1,
                             numRounds=1, numGames=1, numAttempts=1)
        s = ScheduleFactory.createInitialSchedule(conf)
        self.assertTrue(s.isValid())

    def test_factory_simple_bad(self):
        '''
        Bad schedule: 11 players, just 1 game
        '''
        conf = Configuration(numPlayers=11, numTables=1,
                             numRounds=1, numGames=1, numAttempts=1)
        self.assertFalse(conf.isValid())
        s = ScheduleFactory.createInitialSchedule(conf)
        self.assertTrue(s == None)

    def test_factory_rounds2_players10_attempts2(self):
        '''
        Simple 2-round schedule: 10 players 2 attempts per player
        '''
        conf = Configuration(numPlayers=10, numTables=1,
                             numRounds=2, numGames=2, numAttempts=2)
        s = ScheduleFactory.createInitialSchedule(conf)
        self.assertTrue(s.isValid())

    def test_factory_rounds2_players20_attempts1(self):
        '''
        Simple 2-round schedule: 20 players 1 attempts per player
        '''
        conf = Configuration(numPlayers=20, numTables=1,
                             numRounds=2, numGames=2, numAttempts=1)
        s = ScheduleFactory.createInitialSchedule(conf)
        self.assertTrue(s.isValid())

    def test_factory_tables1_players12(self):
        '''
        Classic mini-tournament schedule
        '''
        conf = Configuration(numPlayers=12, numTables=1,
                             numRounds=6, numGames=6, numAttempts=5)
        s = ScheduleFactory.createInitialSchedule(conf)
        self.assertTrue(s.isValid())

    def test_factory_tables2_players20(self):
        '''
        Full schedule for 2 tables
        '''
        conf = Configuration(numPlayers=20, numTables=2,
                             numRounds=10, numGames=20, numAttempts=10)
        s = ScheduleFactory.createInitialSchedule(conf)
        self.assertTrue(s.isValid())

    def test_factory_tables2_players25(self):
        '''
        Not-full schedule for 2 tables (VaWaCa-2017)
        '''
        conf = Configuration(numPlayers=25, numTables=2,
                             numRounds=10, numGames=20, numAttempts=8)
        s = ScheduleFactory.createInitialSchedule(conf)
        self.assertTrue(s.isValid())

    def test_factory_tables3_players30(self):
        '''
        Full schedule for 3 tables
        '''
        conf = Configuration(numPlayers=30, numTables=3,
                             numRounds=10, numGames=30, numAttempts=10)
        s = ScheduleFactory.createInitialSchedule(conf)
        self.assertTrue(s.isValid())

    def test_factory_tables3_players33(self):
        '''
        Not-full schedule for 3 tables (11 rounds by 10 attempts per player)
        '''
        conf = Configuration(numPlayers=33, numTables=3,
                             numRounds=11, numGames=33, numAttempts=10)
        s = ScheduleFactory.createInitialSchedule(conf)
        self.assertTrue(s.isValid())

    def test_factory_tables3_players35(self):
        '''
        Not-full schedule for 3 tables with not all tables at the last round (12 rounds by 10 attempts per player)
        '''
        conf = Configuration(numPlayers=35, numTables=3,
                             numRounds=12, numGames=35, numAttempts=10)
        s = ScheduleFactory.createInitialSchedule(conf)
        self.assertTrue(s.isValid())


if __name__ == '__main__':
    unittest.main()



import json
import os.path

from .schedule import Schedule
from .player import Player, Participants


def saveSchedule(schedule: Schedule, fname: str):
    jsonDict = schedule.toJson()
    jsonStr = json.dumps(jsonDict, indent=2)

    home = os.path.expanduser("~")
    filename = os.path.join(home, fname)
    print(f"Saving schedule to: {filename}")
    with open(filename, "w", encoding="utf-8") as f:
        f.write(jsonStr)


def saveParticipants(participants: Participants, fname: str):
    jsonDict = participants.toJson()
    jsonStr = json.dumps(jsonDict, indent=2)

    home = os.path.expanduser("~")
    filename = os.path.join(home, fname)
    print(f"Saving participants to: {filename}")
    with open(filename, "w", encoding="utf-8") as f:
        f.write(jsonStr)


def loadSchedule(fname: str) -> Schedule:
    home = os.path.expanduser("~")
    filename = os.path.join(home, fname)
    print(f"Loading schedule from: {filename}")
    with open(filename, "r", encoding="utf-8") as f:
        s = f.read()
    d = json.loads(s)

    s = Schedule.fromJson(d)
    s.generateSlotsFromGames()
    return s


def loadParticipants(fname: str) -> Participants:
    home = os.path.expanduser("~")
    filename = os.path.join(home, fname)
    print(f"Loading participants from: {filename}")
    with open(filename, "r", encoding="utf-8") as f:
        s = f.read()
    d = json.loads(s)

    participants = Participants.fromJson(d)
    return participants



import random

from .schedule import Schedule
from .metrics import Metrics
from .game import Game


class OptimizeSeats:
    verbose: bool
    schedule: Schedule

    # this callback is for Schedule that is the best at the current moment
    callbackBetterSchedule: None

    shuffleGameFunc = None

    def log(self, *kargs, **kwargs):
        if self.verbose:
            print(*kargs, **kwargs)

    def __init__(self, schedule: Schedule, verbose: bool = True):
        self.schedule = schedule
        self.verbose = verbose

    def optimize(self, numRuns: int, iterations: list()):
        print("\n*** Optimize seats")

        gamePlayers = self.schedule.saveGamePlayers()
        self.currentScore = None
        self.bestScore = None
        self.bestGamePlayers = None

        for i in range(numRuns):
            print(f"\n*** Seating optimization run: {i+1}")
            self.schedule.updateGamePlayers(gamePlayers)

            func = [
                self.swapAllPlayers,
                self.swapTwoPlayers]

            for stage in range(len(iterations)):
                numIterations = iterations[stage]
                print(f"\nStage: {stage+1} (iterations: {numIterations})")
                self.shuffleGameFunc = func[i % len(func)]
                self.optimizeStage(numIterations)
                if self.bestScore != None and 2 * self.bestScore < self.currentScore:
                    print("We have better best score, so... don't continue")
                    break

            if self.bestGamePlayers == None or self.currentScore < self.bestScore:
                print(f"Found best seating, score : {self.currentScore:8.4f}")
                self.bestScore = self.currentScore
                self.bestGamePlayers = self.schedule.saveGamePlayers()

                if self.callbackBetterSchedule:
                    self.schedule.updateGamePlayers(self.bestGamePlayers)
                    self.callbackBetterSchedule(self.schedule)

        self.schedule.updateGamePlayers(self.bestGamePlayers)

    def optimizeStage(self, iterations: int):
        self.currentScore = self.scoreFunc()

        goodIterations = 0
        for i in range(iterations):
            if i % 1000 == 0:
                print(
                    f"Iteration: {i:8d} of {iterations} (changes: {goodIterations:4d}, score: {self.currentScore:8.4f})")
            success = self.randomSeatChange()
            if success:
                goodIterations += 1

        # debug
        print(f"Final score: {self.currentScore:8.4f}")
        print(f"Good iterations: {goodIterations} of {iterations}")

    def randomSeatChange(self) -> bool:
        game = random.choice(self.schedule.games)

        oldPlayers = game.players.copy()
        self.shuffleGameFunc(game)
        score = self.scoreFunc()

        if score < self.currentScore:
            # debug
            self.log(f"Score: {score:8.4f}. Shuffle game: {game.id}")

            self.currentScore = score
            return True
        else:
            game.players = oldPlayers
            return False

    def swapTwoPlayers(self, game: Game):
        # pick 2 players
        playerOne = random.randrange(len(game.players))
        playerTwo = playerOne
        while playerOne == playerTwo:
            playerTwo = random.randrange(len(game.players))

        # swap two players
        game.players[playerOne], game.players[playerTwo] = game.players[playerTwo], game.players[playerOne]

    def swapAllPlayers(self, game: Game):
        random.shuffle(game.players)

    def scoreFunc(self) -> float:
        m = Metrics(self.schedule)

        target = self.schedule.numAttempts / 10

        # additional heuristics
        calcHalfSimmetry = True
        calcTrippleSimmetry = True
        calcFirstLastSimmetry = True
        factorHalf = 2.0
        factorTripple = 3.0
        factorFirstLast = 20.0

        penalty = 0.0
        for playerId in range(self.schedule.numPlayers):
            seats = m.calcPlayerSeatsHistogram(playerId)
            sd = m.calcSquareDeviation(seats, target)
            penalty += sd

            # half simmetry
            '''if calcHalfSimmetry:
                allSeats = sum(seats)
                k_lo = sum(seats[0:5]) / allSeats
                k_hi = sum(seats[5:10]) / allSeats
                penalty_lo = (k_lo - 0.5) ** 2
                penalty_hi = (k_hi - 0.5) ** 2
                penalty += factorHalf * (penalty_lo + penalty_hi)
            '''

            # tripple simmetry
            if calcTrippleSimmetry:
                allSeats = sum(seats)
                k_a = sum(seats[0:3]) / allSeats
                k_b = sum(seats[3:7]) / allSeats
                k_c = sum(seats[7:10]) / allSeats
                penalty_a = (10 * k_a - 3) ** 2
                penalty_b = (10 * k_b - 4) ** 2
                penalty_c = (10 * k_c - 3) ** 2
                penalty += factorTripple * (penalty_a + penalty_b + penalty_c)

            # first and last simmetry
            if calcFirstLastSimmetry:
                allSeats = sum(seats)
                k_first = seats[0] / allSeats
                k_last = seats[-1] / allSeats
                penalty_first = (10 * k_first - 1) ** 2
                penalty_last = (10 * k_last - 1) ** 2
                penalty += factorFirstLast * (penalty_first + penalty_last)

        return penalty



from dataclasses import dataclass


class ConfigurationException(Exception):
    pass


@dataclass(frozen=True)
class Configuration:
    '''
    Configuration - set of parameters for Mafia tournament
    '''
    # Number of players in tournament
    numPlayers: int

    # Number of tables (games in parallel)
    numTables: int

    # Number of rounds in tournament
    numRounds: int

    # Total number of games: up to  (NumTables * NumRounds)
    numGames: int

    # Total number of games for every player
    numAttempts: int

    def isValid(self) -> bool:
        try:
            self.validate()
        except ConfigurationException:
            return False
        return True

    def validate(self):
        if self.numPlayers < 10:
            raise ConfigurationException(
                f"numPlayers must by >= 10 (value: {self.numPlayers}")

        if self.numTables < 1:
            raise ConfigurationException(
                f"numTables must be >=1 (value: {self.numTables})")

        if self.numRounds < 1:
            raise ConfigurationException(
                f"numRounds must be >=1 (value: {self.numRounds})")

        if self.numGames < 1:
            raise ConfigurationException(
                f"numGames must be >=1 (value: {self.numGames})")

        if self.numAttempts < 1:
            raise ConfigurationException(
                f"numAttempts must be >=1 (value: {self.numAttempts})")

        numGamesLo = self.numTables * (self.numRounds - 1)
        numGamesHi = self.numTables * self.numRounds
        if self.numGames < numGamesLo or self.numGames > numGamesHi:
            raise ConfigurationException(
                f"numGames must be in range [{numGamesLo}, {numGamesHi}] (value: {self.numGames}]")

        if (self.numPlayers * self.numAttempts) % 10 != 0:
            raise ConfigurationException(
                f"numPlayers x numAttempts must be dividable by 10 (value: {self.numPlayers * self.numAttempts})")

        if (self.numPlayers * self.numAttempts) / 10 != self.numGames:
            raise ConfigurationException(
                f"numGames must match numPlayers x numAttempts (numGames: {self.numGames}, numPlayers: {self.numPlayers}, numAttempts: {self.numAttempts}")



import unittest

from configuration import Configuration
from schedule import Schedule
from round import Round
from game import Game


class TestSchedule(unittest.TestCase):
    def test_schedule_CreateEmpty(self):
        s = Schedule(None)
        self.assertEqual(s.configuration, None)
        self.assertEqual(s.participants, None)
        self.assertEqual(s.rounds, [])
        self.assertEqual(s.games, [])

    def test_schedule_CreateSimple(self):
        numPlayers = 10
        conf = Configuration(numPlayers, numTables=1,
                             numRounds=1, numGames=1, numAttempts=1)
        game = Game(1, list(range(numPlayers)))
        games = [game]
        rounds = [Round(1, [game.id])]

        s = Schedule(conf, rounds, games)
        self.assertEqual(s.configuration, conf)
        self.assertEqual(s.participants, None)
        self.assertEqual(len(s.rounds), 1)
        self.assertEqual(len(s.games), 1)

    pass


if __name__ == '__main__':
    unittest.main()



import dataclasses
import unittest

from player import Player


class TestPlayer(unittest.TestCase):
    def setUp(self):
        # print("Before test")
        pass

    def tearDown(self):
        # print("After test")
        pass

    def test_player_Create(self):
        player = Player(1, 'Nickname')
        self.assertEqual(player.id, 1)
        self.assertEqual(player.name, 'Nickname')

    def test_player_CreateIntName(self):
        '''
        Unfortunately, it is still possible to create player with integer (non string!) name
        '''
        player = Player(147, 42)
        self.assertEqual(player.id, 147)
        self.assertEqual(player.name, 42)

    def test_player_IdReadonly(self):
        player = Player(1, 'name')
        self.assertEqual(player.id, 1)
        with self.assertRaises(dataclasses.FrozenInstanceError):
            player.id = 2

    def test_player_NameReadonly(self):
        player = Player(1, 'docent')
        self.assertEqual(player.name, 'docent')
        with self.assertRaises(dataclasses.FrozenInstanceError):
            player.name = 'like'

    pass


if __name__ == '__main__':
    unittest.main()



import dataclasses
import unittest

from configuration import Configuration, ConfigurationException


class TestConfiguration(unittest.TestCase):
    def test_configuration_CreateSimple(self):
        c = Configuration(33, 3, 11, 33, 10)
        self.assertEqual(c.numPlayers, 33)
        self.assertEqual(c.numTables, 3)
        self.assertEqual(c.numRounds, 11)
        self.assertEqual(c.numGames, 33)
        self.assertEqual(c.numAttempts, 10)

    def test_configuration_Readonly(self):
        c = Configuration(30, 3, 10, 30, 10)
        with self.assertRaises(dataclasses.FrozenInstanceError):
            c.numPlayers = 40
        with self.assertRaises(dataclasses.FrozenInstanceError):
            c.numTables = 20
        with self.assertRaises(dataclasses.FrozenInstanceError):
            c.numRounds = 12
        with self.assertRaises(dataclasses.FrozenInstanceError):
            c.numGames = 8
        with self.assertRaises(dataclasses.FrozenInstanceError):
            c.numAttempts = 1

    def test_configuration_validate_oneTable(self):
        c = Configuration(numPlayers=12, numTables=1,
                          numRounds=6, numGames=6, numAttempts=5)
        self.assertTrue(c.isValid())

        c = Configuration(numPlayers=10, numTables=1,
                          numRounds=3, numGames=3, numAttempts=3)
        self.assertTrue(c.isValid())

    def test_configuration_validate_twoTables(self):
        c = Configuration(numPlayers=20, numTables=2,
                          numRounds=10, numGames=20, numAttempts=10)
        self.assertTrue(c.isValid())

        c = Configuration(numPlayers=25, numTables=2,
                          numRounds=10, numGames=20, numAttempts=8)
        self.assertTrue(c.isValid())

    def test_configuration_validate_threeTables(self):
        c = Configuration(numPlayers=30, numTables=3,
                          numRounds=10, numGames=30, numAttempts=10)
        self.assertTrue(c.isValid())

        c = Configuration(numPlayers=36, numTables=3,
                          numRounds=12, numGames=36, numAttempts=10)
        self.assertTrue(c.isValid())

        c = Configuration(numPlayers=37, numTables=3,
                          numRounds=13, numGames=37, numAttempts=10)
        self.assertTrue(c.isValid())

    def test_configuration_validate_fail_players(self):
        '''players must be >=10'''
        with self.assertRaises(ConfigurationException):
            c = Configuration(numPlayers=-1, numTables=1,
                              numRounds=10, numGames=10, numAttempts=10)
            c.validate()

        with self.assertRaises(ConfigurationException):
            c = Configuration(numPlayers=9, numTables=1,
                              numRounds=10, numGames=10, numAttempts=10)
            c.validate()

    def test_configuration_validate_fail_zeroValues(self):
        with self.assertRaises(ConfigurationException):
            c = Configuration(numPlayers=0, numTables=1,
                              numRounds=10, numGames=10, numAttempts=10)
            c.validate()

        with self.assertRaises(ConfigurationException):
            c = Configuration(numPlayers=10, numTables=0,
                              numRounds=10, numGames=10, numAttempts=10)
            c.validate()

        with self.assertRaises(ConfigurationException):
            c = Configuration(numPlayers=10, numTables=1,
                              numRounds=0, numGames=10, numAttempts=10)
            c.validate()

        with self.assertRaises(ConfigurationException):
            c = Configuration(numPlayers=10, numTables=1,
                              numRounds=10, numGames=0, numAttempts=10)
            c.validate()

        with self.assertRaises(ConfigurationException):
            c = Configuration(numPlayers=10, numTables=1,
                              numRounds=10, numGames=10, numAttempts=0)
            c.validate()

    def test_configuration_validate_gamesRoundsMatch(self):
        '''games must be in range [tables * rounds-1; rables * rounds]'''
        with self.assertRaises(ConfigurationException):
            # rounds must be 12 here
            c = Configuration(numPlayers=35, numTables=3,
                              numRounds=10, numGames=35, numAttempts=10)
            c.validate()

        with self.assertRaises(ConfigurationException):
            # rounds must be 12 here
            c = Configuration(numPlayers=35, numTables=3,
                              numRounds=13, numGames=35, numAttempts=10)
            c.validate()

    def test_configuration_validate_attemptsGamesMatch(self):
        '''players x attempts must be == games * 10'''
        with self.assertRaises(ConfigurationException):
            # attempts must be 10 here
            c = Configuration(numPlayers=35, numTables=3,
                              numRounds=12, numGames=35, numAttempts=8)
            c.validate()

        with self.assertRaises(ConfigurationException):
            # attempts must be 10 here
            c = Configuration(numPlayers=35, numTables=3,
                              numRounds=12, numGames=35, numAttempts=12)
            c.validate()


if __name__ == '__main__':
    unittest.main()



import dataclasses


@dataclasses.dataclass(frozen=True)
class Player:
    '''Single player of torunament. Player has unique <id> and <name>'''

    id: int
    name: str
    pass


@dataclasses.dataclass(order=False, )
class Participants:
    ''' Participants of tournament. Automatically assigns ids to players in constructor'''

    people: list[Player] = dataclasses.field(default_factory=list)

    # don't declare as it will be included into JSON
    #_peopleDict: dict[int, Player] = None

    def __len__(self):
        return len(self.people)

    def __getitem__(self, index):
        return self.people[index]

    def find(self, id) -> Player:
        return self._peopleDict.get(id)

    def __post_init__(self):
        self._peopleDict = {}
        for player in self.people:
            if player.id in self._peopleDict:
                raise RuntimeError(
                    f'Can not create participants: non-unique player id: {player.id}')
            self._peopleDict[player.id] = player

    @staticmethod
    def create(count: int, prefix: str = None):
        names = Participants.generateNames(count, prefix)
        return Participants.createFromNames(names)

    @staticmethod
    def createFromNames(names: list):
        people = []
        id = 0
        for name in names:
            # create a player and add to dict
            player = Player(id, name)
            id = id + 1
            people.append(player)

        return Participants(people)

    @staticmethod
    def generateNames(count: int, prefix: str = None):
        s = prefix if prefix is not None else 'p'

        names = []
        for id in range(count):
            name = f"{s}{id:02d}"
            names.append(name)
        return names

    def toJson(self):
        return dataclasses.asdict(self)

    @staticmethod
    def fromJson(d: dict):
        p = [Player(**p) for p in d['people']]
        return Participants(p)

    pass



import math

from .game import Game
from .player import Player
from .schedule import Schedule


class Metrics:
    def __init__(self, schedule: Schedule):
        self.schedule = schedule
        self.idealHist = None

    def calcPlayerOpponents(self, thisPlayerId: int):
        '''
        Calculates opponents histogram for given <thisPlayerId>
        Returns a list of size <numPlayers>,
        where value of a[i] - total number of games that pair of players <thisPlayerId>, <i> are playing
        '''
        opponents = [0] * self.schedule.numPlayers

        # go through all games
        for slotId, slot in self.schedule.slots.items():
            if thisPlayerId in slot.players:
                for id in slot.players:
                    if thisPlayerId != id:
                        opponents[id] += 1

        return opponents

    def calcOpponentsMatrix(self):
        '''Calculates opponents histogram for all players'''
        matrix = []
        for playerId in range(self.schedule.numPlayers):
            line = self.calcPlayerOpponents(playerId)
            matrix.append(line)
        return matrix

    def calcPlayerPairsHistogram(self, thisPlayerId: int):
        '''
        Calculates pairs histogram for current player.
        Histogram is dictionary <numGames>:<numPairs>
        '''
        pairs = {}
        for val in range(self.schedule.numAttempts + 1):
            pairs[val] = 0

        opponents = self.calcPlayerOpponents(thisPlayerId)
        for id in range(len(opponents)):
            if id != thisPlayerId:
                numGames = opponents[id]
                pairs[numGames] += 1
        return pairs

    def calcPlayerPairs(self, thisPlayerId: int):
        '''
        Calculates pairs for current user.
        This is dictionary <numGames>:[opponents]
        '''
        pairs = {}
        for numGames in range(0, self.schedule.numAttempts + 1):
            pairs[numGames] = []

        opponents = self.calcPlayerOpponents(thisPlayerId)
        for id in range(len(opponents)):
            if id != thisPlayerId:
                numGames = opponents[id]
                pairs[numGames].append(id)
        return pairs

    def calcPairsHistogram(self):
        '''
        Calculates pairs histogram for all players.
        Histogram is dictionary <numGames> : <numPairs>
        '''
        allPairs = {}
        for val in range(self.schedule.numAttempts + 1):
            allPairs[val] = 0

        for playerId in range(self.schedule.numPlayers):
            pairs = self.calcPlayerPairsHistogram(playerId)

            for numGames, numPairs in pairs.items():
                allPairs[numGames] += numPairs
        return allPairs

    def pairsHistogramSum(self, pairs: dict):
        result = 0
        for numGames, numPairs in pairs.items():
            result += numPairs
        return result

    def pairsHistogramCenter(self, pairs: dict):
        totalMult = 0
        totalPairs = 0
        for numGames, numPairs in pairs.items():
            totalMult += (numGames + 1) * numPairs * numPairs
            totalPairs += numPairs * numPairs

        center = totalMult / totalPairs
        return center

    def pairsHistogramRange(self, pairs: dict):
        beg = None
        for idx in sorted(pairs):
            if beg == None and pairs[idx] > 0:
                beg = idx

        end = None
        for idx in reversed(sorted(pairs)):
            if end == None and pairs[idx] > 0:
                end = idx
        return beg, end

    def penaltyIdealHistogram(self):
        hist = {}
        for val in range(self.schedule.numAttempts + 1):
            hist[val] = 0

        numPairs = self.schedule.numPlayers - 1
        target = 9 * self.schedule.numAttempts / (self.schedule.numPlayers-1)

        idxLow = math.floor(target)
        idxHi = math.ceil(target)
        hist[idxLow] = numPairs * (idxHi - target)
        hist[idxHi] = numPairs - hist[idxLow]
        return hist

    def penaltyPlayer(self, playerId: int):
        target = 9 * self.schedule.numAttempts / (self.schedule.numPlayers-1)
        if self.idealHist == None:
            self.idealHist = self.penaltyIdealHistogram()

        hist = self.calcPlayerPairsHistogram(playerId)

        penalty = 0.0
        for idx in hist:
            idxDist = (idx - target) ** 2
            valueDist = abs(self.idealHist[idx] - hist[idx])  # ** 2
            penalty += idxDist * valueDist
        return penalty

    def calcPlayerSeatsHistogram(self, thisPlayerId: int):
        '''
        Calculates seats histogram for given <thisPlayerId>
        Returns a list of size 10,
        where value of a[i] - total number of games where player sits on seat number <i>
        '''
        seats = [0] * 10

        # go through all games
        for game in self.schedule.games:
            for seat_idx in range(0, 10):
                if game.players[seat_idx] == thisPlayerId:
                    seats[seat_idx] += 1
        return seats

    def calcSeatsMatrix(self):
        '''Calculates seats histogram for all players'''
        matrix = []
        for playerId in range(self.schedule.numPlayers):
            line = self.calcPlayerSeatsHistogram(playerId)
            matrix.append(line)
        return matrix

    def calcSquareDeviationExclude(self, data: list, target: float, exclude_idx: int):
        '''
        Calculates standard deviation between values in <data> and <target>, 
        except for item data[exclude_idx].
        '''
        sd = -(data[exclude_idx] - target) * (data[exclude_idx] - target)
        for i in range(len(data)):
            sd += (data[i] - target) * (data[i] - target)
        return sd / (len(data) - 1)

    def calcSquareDeviation(self, data: list, target: float):
        '''
        Calculates standard deviation between values in <data> and <target> value.
        '''
        sd = 0.0
        for i in range(len(data)):
            sd += (data[i] - target) * (data[i] - target)
        return sd / len(data)



import random

from .configuration import Configuration
from .schedule_factory import ScheduleFactory
from .schedule import Schedule
from .metrics import Metrics


class OptimizeOpponents:
    verbose: bool

    # current score and schedule
    schedule: Schedule
    score: float

    # best score and schedule
    bestSchedule: Schedule
    bestScore: float

    # TODO: refactor into expectedPairs list, -1 means no constraint
    expectedZeroPairs: int
    expectedSinglePairs: int

    # this callback is for every Schedule generated on iteration if it is not that good
    callbackCurrSchedule = None

    # this callback is for Schedule that is the best at the current moment
    callbackBetterSchedule = None

    def log(self, *kargs, **kwargs):
        if self.verbose:
            print(*kargs, **kwargs)

    def __init__(self, verbose: bool = True):
        self.verbose = verbose
        # self.expectedZeroPairs = expectedZeroPairs

    def optimize(self, conf: Configuration, numRuns: int, numIterations: int):
        print("\n*** Optimize opponents")

        self.bestSchedule = None
        self.bestScore = 0
        for i in range(numRuns):
            print(f"\n*** Opponents optimization run: {i+1}")
            self.schedule = ScheduleFactory.createInitialSchedule(conf)
            self.schedule.generateSlotsFromGames()
            self.score = self.scoreFunc()
            self.optimizeStage(numIterations)

            self.schedule.updateGamesFromSlots()

            if not self.bestSchedule or self.score < self.bestScore:
                self.bestSchedule = self.schedule
                self.bestScore = self.score

                # callback for better schedule
                if self.callbackBetterSchedule:
                    self.callbackBetterSchedule(self.bestSchedule)
            else:
                # callback for every schedule
                if self.callbackCurrSchedule:
                    self.callbackCurrSchedule(self.schedule)

            self.schedule = None
            self.score = 0

        return self.bestSchedule

    def optimizeStage(self, numIterations: int):
        goodIterations = 0
        for i in range(0, numIterations):
            # debug
            if i % 1000 == 0:
                print(
                    f"Iteration: {i:8d} of {numIterations} (changes: {goodIterations:4d}, score: {self.score:8.4f})")

            success = self.randomOpponentChange()
            if success:
                goodIterations += 1

        # debug
        print(f"Final score: {self.score:8.4f}")
        print(f"Good iterations: {goodIterations} of {numIterations}")

    def randomOpponentChange(self) -> bool:
        if self.schedule.configuration.numTables == 1:
            roundOne = random.choice(self.schedule.rounds)
            roundTwo = roundOne
            while roundTwo == roundOne:
                roundTwo = random.choice(self.schedule.rounds)
            return self.randomOpponentChangeInRounds(roundOne.id, roundTwo.id)

        # special case - last round is NOT full and only contains 1 table
        # so we back up to
        r = None
        while (r is None) or (len(r.gameIds) == 1):
            r = random.choice(self.schedule.rounds)

        gameOneId = random.choice(r.gameIds)
        gameTwoId = gameOneId
        while gameTwoId == gameOneId:
            gameTwoId = random.choice(r.gameIds)
        return self.randomOpponentChangeInGames(gameOneId, gameTwoId)

    def randomOpponentChangeInRounds(self, roundOneId: int, roundTwoId: int) -> bool:
        # in this special case every round has one and only one game
        # that's why game index is equal to round index
        gameOneId = self.schedule.games[roundOneId].id
        gameTwoId = self.schedule.games[roundTwoId].id

        slotOne = self.schedule.slots[gameOneId]
        slotTwo = self.schedule.slots[gameTwoId]

        # figure out what players can be switched
        all = {playerId for playerId in range(self.schedule.numPlayers)}
        busyOne = slotOne.players
        freeOne = all.difference(busyOne)
        busyTwo = slotTwo.players
        freeTwo = all.difference(busyTwo)

        # chose 2 players to switch between games
        poolA = busyOne.intersection(freeTwo)
        poolB = busyTwo.intersection(freeOne)
        if len(poolA) == 0 or len(poolB) == 0:
            # can not find substitution as one of player pools is empty
            # print("empty pool!")
            return False

        playerA = random.choice(list(poolA))
        playerB = random.choice(list(poolB))

        # switch players from games
        slotOne.players.remove(playerA)
        slotOne.players.add(playerB)
        slotTwo.players.remove(playerB)
        slotTwo.players.add(playerA)

        # continue only if score gets better
        currentScore = self.scoreFunc()
        if currentScore < self.score:
            self.score = currentScore
            self.log(
                f"Score: {self.score:8.4f}. " +
                f"Swap in rounds: {roundOneId:2d} x {roundTwoId:2d}, players: {playerA:2d} x {playerB:2d}")
            return True
        else:
            slotOne.players.remove(playerB)
            slotOne.players.add(playerA)
            slotTwo.players.remove(playerA)
            slotTwo.players.add(playerB)
            return False

    def randomOpponentChangeInGames(self, gameOneId: int, gameTwoId: int) -> bool:
        slotOne = self.schedule.slots[gameOneId]
        slotTwo = self.schedule.slots[gameTwoId]

        busyOne = slotOne.players
        busyTwo = slotTwo.players
        busyBoth = busyOne.intersection(busyTwo)

        one = busyOne.difference(busyBoth)
        two = busyTwo.difference(busyBoth)
        if len(one) == 0 or len(two) == 0:
            # no candidates to swap
            # print("No candidates to swap!")
            return False

        playerA = random.choice(list(one))
        playerB = random.choice(list(two))

        # switch players from games
        slotOne.players.remove(playerA)
        slotOne.players.add(playerB)
        slotTwo.players.remove(playerB)
        slotTwo.players.add(playerA)

        # continue only if score gets better
        currentScore = self.scoreFunc()
        if currentScore < self.score:
            self.score = currentScore
            self.log(f"Score: {self.score:8.4f}. " +
                     f"Swap in games: {gameOneId:2d} x {gameTwoId:2d}, players: {playerA:2d} x {playerB:2d}")
            return True
        else:
            slotOne.players.remove(playerB)
            slotOne.players.add(playerA)
            slotTwo.players.remove(playerA)
            slotTwo.players.add(playerB)
            return False

    '''
    # old score func
    def scoreFunc(self) -> float:
        metrics = Metrics(self.schedule)

        target = 9 * self.schedule.numAttempts / (self.schedule.numPlayers-1)
        penalty = 0.0
        for playerId in range(self.schedule.numPlayers):
            opponents = metrics.calcPlayerOpponents(playerId)
            sd = metrics.calcSquareDeviationExclude(
                opponents, target, playerId)
            penalty += sd
        return penalty
    '''

    # new score function
    def scoreFunc(self) -> float:
        metrics = Metrics(self.schedule)

        basePenalty = 0.0
        for playerId in range(self.schedule.numPlayers):
            basePenalty += metrics.penaltyPlayer(playerId)

        zeroPlayers = 0
        singlePlayers = 0
        for playerId in range(self.schedule.numPlayers):
            pairs = metrics.calcPlayerPairsHistogram(playerId)
            if pairs[0] == 1:
                zeroPlayers += 1
            if pairs[1] == 1:
                singlePlayers += 1

        # big number to affect basePenalty
        penaltyCoefficient = 100

        zeroPenalty = 0
        if self.expectedZeroPairs > 0:
            expectedZeroPlayers = 2 * self.expectedZeroPairs
            zeroPenalty = penaltyCoefficient * \
                (zeroPlayers - expectedZeroPlayers) ** 2

        singlePenalty = 0
        if self.expectedSinglePairs > 0:
            expectedSinglePlayers = 2 * self.expectedSinglePairs
            singlePenalty = penaltyCoefficient * \
                (singlePlayers - expectedSinglePlayers) ** 2

        return basePenalty + zeroPenalty + singlePenalty



import unittest

from game import Game
from player import Player, Participants


class TestGame(unittest.TestCase):
    def test_game_CreateEmpty(self):
        game = Game(7, [])
        self.assertEqual(game.id, 7)
        self.assertEqual(len(game.players), 0)
        self.assertFalse(game.isValid())

    def test_game_CreateSimple(self):
        participants = Participants.create(10)
        players = participants.people
        game = Game(1, players)
        self.assertEqual(len(game.players), 10)
        self.assertTrue(game.isValid())

    def test_game_isValid_WrongCount(self):
        participants = Participants.create(10)
        players = participants.people
        players = players[:-1]
        game = Game(1, players)

        self.assertEqual(len(game.players), 9)
        self.assertFalse(game.isValid())

    def test_game_isValid_NonUnique(self):
        # duplicate the fifth player
        participants = Participants.create(10)
        players = participants.people
        players[7] = players[5]

        game = Game(1, players)
        self.assertEqual(len(game.players), 10)
        self.assertFalse(game.isValid())

    pass


if __name__ == '__main__':
    unittest.main()



import dataclasses

from .configuration import *
from .round import *
from .game import *
from .player import *


class ScheduleException(Exception):
    pass


class Schedule:
    _configuration: Configuration
    _participants: Participants
    _rounds = list[Round]
    _games = list[Game]

    @property
    def configuration(self) -> Configuration:
        return self._configuration

    @property
    def participants(self) -> Participants:
        return self._participants

    @property
    def rounds(self) -> list[Round]:
        return self._rounds

    @property
    def games(self) -> list[Game]:
        return self._games

    @property
    def numPlayers(self) -> int:
        return self._configuration.numPlayers

    @property
    def numTables(self) -> int:
        return self._configuration.numTables

    @property
    def numRounds(self) -> int:
        return self._configuration.numRounds

    @property
    def numGames(self) -> int:
        return self._configuration.numGames

    @property
    def numAttempts(self) -> int:
        return self._configuration.numAttempts

    def __init__(self, conf: Configuration = None, rounds: list = [], games: list = []):
        self._configuration = conf
        self._participants = None
        self._rounds = rounds
        self._games = games

    def setParticipants(self, people: Participants):
        self._participants = people

    def toJson(self):
        d = {}
        d['configuration'] = dataclasses.asdict(self._configuration)
        d['rounds'] = [dataclasses.asdict(round) for round in self._rounds]
        d['games'] = [dataclasses.asdict(game) for game in self._games]
        return d

    @staticmethod
    def fromJson(d: dict):
        conf = Configuration(**d['configuration'])
        rounds = [Round.fromJson(item) for item in d['rounds']]
        games = [Game.fromJson(item) for item in d['games']]
        return Schedule(conf, rounds, games)

    def generateSlotsFromGames(self):
        self.slots = {}
        for game in self.games:
            self.slots[game.id] = GameSet.create(game)

    def updateGamesFromSlots(self):
        for slotId in self.slots:
            self.games[slotId] = Game.create(self.slots[slotId])
        # Careful. Now we have both games and slots and they may not match each other...
        # self.slots = {}

    def saveGamePlayers(self) -> dict[int, list[int]]:
        '''
        Saves players of all games to dictionary.
        This function is used by multi-run seats optimization.
        '''
        gamePlayers = {}
        for game in self.games:
            gamePlayers[game.id] = list(game.players)
        return gamePlayers

    def updateGamePlayers(self, gamePlayers: dict[int, list[int]]):
        '''
        Loads game players from external dictionary.
        This function is used by multi-run seats optimization.
        '''
        for game in self.games:
            if game.id not in gamePlayers.keys():
                raise RuntimeError(
                    f"UpdatePlayers. Game: {game.id} is not specified in gamePlayers!")
            players = gamePlayers[game.id]
            game.players = list(players)

    def isValid(self) -> bool:
        try:
            self.validate()
        except ScheduleException:
            return False
        return True

    def validateGame(self, game):
        players = set[int]()
        for playerId in game.players:
            if playerId < 0 or playerId > self.numPlayers:
                raise ScheduleException(f"Invalid player id: {player.id}")
            if playerId in players:
                raise ScheduleException(
                    f"Player: {playerId} can not play in a single game #{game.id} twice!")
            players.add(playerId)

    def validateRound(self, round):
        players = set[int]()
        for gameId in round.gameIds:
            game = self.games[gameId]
            for playerId in game.players:
                if playerId in players:
                    raise ScheduleException(
                        f"Player: {playerId} can not play in a single round #{round.id} twice!")
                players.add(playerId)

    def validate(self):
        if self._participants != None and len(self._participants) != self.numPlayers:
            raise ScheduleException(
                f"Participant count: {self._players} must match configuration: {self.numPlayers}")

        if len(self._rounds) != self.numRounds:
            raise ScheduleException(
                f"Round count: {self._rounds} must match configuration: {self.numRounds}")

        if len(self._games) != self.numGames:
            raise ScheduleException(
                f"Game count: {self._games} must match configuration: {self.numGames}")

        # check game count in rounds
        for i, round in enumerate(self._rounds):
            fullRoundGames = self.numTables
            lastRoundGames = self.numGames - \
                (self.numRounds - 1) * self.numTables
            if i < len(self.rounds) - 1:
                if len(round.gameIds) != fullRoundGames:
                    raise ScheduleException(
                        f"Wrong game count in round: {i}. Expected: {fullRoundGames}, got: {len(round.gameIds)}")
            else:
                if len(round.gameIds) != lastRoundGames:
                    raise ScheduleException(
                        f"Wrong game count in last round: {id}. Expected: {lastRoundGames}, got: {len(round.gameIds)}")

        # check that rounds have all games in games
        # check every game is in one and only one round
        allIds = {game.id for game in self._games}
        leftIds = allIds.copy()
        for round in self._rounds:
            for gameId in round.gameIds:
                if gameId not in allIds:
                    raise ScheduleException(
                        f"Wrong game id: {gameId} in round: {round.id}")
                if gameId not in leftIds:
                    raise ScheduleException(
                        f"Duplicate game id: {gameId} in round: {round.id}")
                else:
                    leftIds.remove(gameId)
        if len(leftIds) != 0:
            raise ScheduleException(
                f"Some games are not in any round: {leftIds}")

        # validate game
        for game in self._games:
            self.validateGame(game)

        # validate round
        for round in self._rounds:
            self.validateRound(round)

        # calc number of games played by every player
        gamesPlayed = {}
        for game in self._games:
            for playerId in game.players:
                num = 1 + gamesPlayed.get(playerId, 0)
                gamesPlayed[playerId] = num

        # all players must play the same number of attempts
        counts = gamesPlayed.values()
        gamesPlayedSet = set(counts)
        if len(gamesPlayedSet) != 1:
            raise ScheduleException(
                f"All players must play the same number of games")

        # every player must play NumAttempts games
        for playerId in range(self.numPlayers):
            if playerId not in gamesPlayed:
                raise ScheduleException(
                    f"Player: {playerId} does not play a single game!")

            gamesPlayedById = gamesPlayed[playerId]
            if gamesPlayed[playerId] != self.numAttempts:
                raise ScheduleException(
                    f"Player: {player.id} game count: {gamesPlayedById} must match configuration: {self.numAttempts}")



'''
The module contains set of classes and methods to generate schedule for mafia tournaments.
Also, it contains utils to work with schedules
'''

from .configuration import Configuration, ConfigurationException
from .schedule import Schedule, ScheduleException
from .schedule_factory import ScheduleFactory

from .game import Game
from .round import Round
from .player import Player

from .metrics import Metrics
from .optimize_seats import OptimizeSeats
from .optimize_opponents import OptimizeOpponents

from .output import Print
# from .mafia_schedule.helpers import *
